from __future__ import annotations

from io import StringIO

import pytest

from abSENSE.parameters import AbsenseParameters
from abSENSE.recorder import Recorder


@pytest.fixture()
def faked_recorder(mocker, default_params):
    mocker.patch("abSENSE.recorder.os.makedirs")
    default_params.out_dir = "."
    default_params.predict_all = True
    recorder = Recorder.build_recorder(default_params, ["s1", "s2", "s3"])
    # replace files with stringIO for easier testing
    files = {}

    def new_stringIO(*args, **kwargs):
        result = StringIO()
        result.name = args[0]
        files[args[0]] = result
        return result

    mocker.patch("abSENSE.recorder.open", side_effect=new_stringIO)
    return recorder, files


@pytest.fixture()
def faked_recorder_non_closing(mocker, default_params):
    mocker.patch("abSENSE.recorder.os.makedirs")
    default_params.out_dir = "."
    recorder = Recorder.build_recorder(default_params, ["s1", "s2", "s3"])
    # replace files with stringIO for easier testing
    files = {}

    def new_stringIO(*args, **kwargs):
        result = StringIO()
        result.close = lambda: None
        result.name = args[0]
        files[args[0]] = result
        return result

    mocker.patch("abSENSE.recorder.open", side_effect=new_stringIO)
    return recorder, files


@pytest.fixture()
def quick_bitscores():
    return (
        "Gene\tS_cer\tS_par\tS_mik\tS_bay\tS_kud\tS_castellii\tK_waltii\tK_lactis\tA_gossyppi\tY_lipolytica\tA_nidulans\tS_pombe\n"
        "NP_010181.2\t2284.0\t2254.0\t2234.0\t2200.0\t2197.0\t1895.0\t1767.0\t1691.0\t1736.0\t1297.0\t1099.0\t1194.0\n"
        "NP_009362.1\t1027.0\t1016.0\t1018.0\t1008.0\t1010.0\t909.0\t903.0\t894.0\t875.0\t714.0\t691.0\t684.0\n"
        "NP_014555.1\t712.0\t698.0\t696.0\t684.0\t688.0\tN/A\t617.0\t611.0\tN/A\tN/A\tN/A\t344.0\n"
        "NP_116682.3\t352.0\t171.0\t92.8\t0\t82.0\t0\t0\t0\t0\t0\t0\t0\n"
        "NP_011284.1\t308.0\t225.0\t179.0\t191.0\t178.0\t0\t0\t0\t0\t0\t0\t0\n"
        "NP_011878.1\t588.0\t522.0\t446.0\t439.0\t455.0\t0\t0\t0\t0\t0\t0\t0\n"
        "NP_013320.1\t1491.0\t1205.0\t1013.0\t910.0\t996.0\t141.0\t101.0\t102.0\t63.2\t0\t0\t0\n"
        "NP_014160.2\t941.0\t927.0\t889.0\t885.0\t893.0\t638.0\t0\t0\t657.0\t375.0\t378.0\t0\n"
        "NP_014890.1\t394.0\t264.0\t254.0\t230.0\t271.0\t58.9\t0\t0\t0\t0\t0\t0\n"
    )


@pytest.fixture()
def quick_distances():
    return (
        "#Species\tDistance\n"
        "S_cer\t0\n"
        "S_par\t0.051\n"
        "S_mik\t0.088\n"
        "S_kud\t0.104\n"
        "S_bay\t0.108\n"
        "S_castellii\t0.363\n"
        "K_waltii\t0.494\n"
        "A_gossyppi\t0.518\n"
        "K_lactis\t0.557\n"
        "A_nidulans\t0.903\n"
        "S_pombe\t0.922\n"
        "Y_lipolytica\t0.954\n"
    )


@pytest.fixture()
def quick_distances_replicates():
    return (
        "S_cer\t0\t0\t0\t0\n"
        "S_par\t0.051\t0.061\t0.041\t0.052\n"
        "S_mik\t0.088\t0.098\t0.078\t0.098\n"
        "S_kud\t0.104\t0.114\t0.094\t0.114\n"
        "S_bay\t0.108\t0.118\t0.098\t0.107\n"
        "S_castellii\t0.363\t0.353\t0.373\t0.366\n"
        "K_waltii\t0.494\t0.484\t0.504\t0.496\n"
        "A_gossyppi\t0.518\t0.508\t0.528\t0.520\n"
        "K_lactis\t0.557\t0.547\t0.567\t0.559\n"
        "A_nidulans\t0.903\t0.893\t0.913\t0.905\n"
        "S_pombe\t0.922\t0.912\t0.932\t0.924\n"
        "Y_lipolytica\t0.954\t0.944\t0.964\t0.956\n"
    )


@pytest.fixture()
def fungi_database_lengths():
    return (
        "#Species\tDBsize\n"
        "A_nidulans\t5073867\n"
        "Y_lipolytica\t3139837\n"
        "A_gossyppi\t2335136\n"
        "K_lactis\t2462503\n"
        "S_pombe\t2381887\n"
        "S_castellii\t2768407\n"
        "S_mik\t2612415\n"
        "S_par\t2606124\n"
        "S_kud\t2603233\n"
        "S_cer\t2615464\n"
        "S_bay\t2595487\n"
        "K_waltii\t2424992\n"
    )


@pytest.fixture()
def fungi_gene_lengths():
    return (
        "#GeneID\tGeneLength\n"
        "NP_001018029.1\t66\n"
        "NP_001018030.1\t360\n"
        "NP_001018031.2\t113\n"
        "NP_001018032.1\t52\n"
        "NP_001018033.3\t92\n"
        "NP_001027023.1\t163\n"
        "NP_001027534.1\t73\n"
        "NP_001032571.3\t67\n"
        "NP_001032572.1\t78\n"
        "NP_001032573.1\t79\n"
        "NP_010181.2\t101\n"
        "NP_009362.1\t102\n"
        "NP_014555.1\t103\n"
        "NP_116682.3\t104\n"
        "NP_011284.1\t105\n"
        "NP_011878.1\t106\n"
        "NP_013320.1\t107\n"
        "NP_014160.2\t108\n"
        "NP_014890.1\t109\n"
    )


@pytest.fixture
def default_params(quick_bitscores, quick_distances):
    distances = StringIO(quick_distances)
    distances.name = "distances"
    bitscores = StringIO(quick_bitscores)
    bitscores.name = "bitscores"
    return AbsenseParameters(
        distances=distances,
        bitscores=bitscores,
        e_value=0.001,
        predict_all=False,
        plot_all=False,
        validate=False,
        plot="",
        include_only=None,
        gene_lengths=None,
        db_lengths=None,
        out_dir=None,
        start_time="NOW",
    )



"""Wrapper for parameters from click."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, TextIO


@dataclass
class AbsenseParameters:
    """Collection of parameters from the command line."""

    bitscores: TextIO
    db_lengths: TextIO | None
    distances: TextIO
    e_value: float
    gene_lengths: TextIO | None
    include_only: str | None
    out_dir: str | None
    plot: str
    plot_all: bool
    predict_all: bool
    start_time: str
    validate: bool
    default_gene_length: float = 400
    default_db_length: float = 8_000_000

    @property
    def output_directory(self) -> str:
        """Get the output directory supplied or the default."""
        if self.out_dir is None:
            return f"abSENSE_results_{self.start_time}"

        return self.out_dir

    def need_plots(self) -> bool:
        """Return true if the params require plotting."""
        return self.plot_all or self.plot != ""

    def plot_test(self) -> Callable[[str], bool]:
        """Return a function to test if a gene should be plotted."""
        targets = set(self.plot.split(","))

        def should_plot(gene: str) -> bool:
            if self.plot_all:
                return True
            return gene in targets

        return should_plot



"""Module for producing plots for abSENSE results."""
from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from matplotlib.patches import Patch

from abSENSE.constants import GREY, ORANGE
from abSENSE.utilities import exponential, find_confidence_interval, sample_parameters


class FitPlot:
    """Wrapper of matplotlib to specialize for abSENSEE."""

    def __init__(self) -> None:
        """Initialize new plot."""
        self.figure, self.axes = plt.subplots(figsize=(11.5, 6.0), tight_layout=True)
        self.interval = 0.01

    def title(self, gene: str, a_fit: float, b_fit: float, correlation: float) -> None:
        """Place title on plot.

        Args:
            gene: gene name
            a_fit: a fit value in exponential function
            b_fit: b fit value in exponential function
            correlation: goodness of fit
        """
        self.axes.set_title(
            (
                f"Gene: {gene}\n"
                f"a = {round(a_fit, 1)}, b = {round(b_fit, 2)}\n"
                f"$r^2$ = {round(correlation**2, 2)}"
            ),
            color="black",
            fontsize=13,
            fontweight="bold",
        )

    def scores(
        self,
        distance: list[float],
        score: list[float],
        xerr: list[float],
    ) -> None:
        """Plot score values.

        Args:
            distance: x values
            score: y values
            xerr: uncertainty in x values, None to skip plotting errorbar
        """
        # passing in na values raises a runtime error for errorbar
        self.axes.errorbar(
            distance,
            np.zeros_like(distance),
            xerr=xerr,
            ecolor="red",
            fmt="none",
            elinewidth=4,
        )
        self.axes.scatter(
            x=distance,
            y=score,
            s=40,
            c="black",
            label="Bitscores of detected orthologs used in fit",
        )

    def fit(
        self,
        distance: npt.NDArray[np.float64],
        a_fit: float,
        b_fit: float,
        covariance: npt.NDArray[np.float64],
        bit_threshold: float,
    ) -> None:
        """Plot fit and confidence interval.

        Args:
            distance: target x values
            a_fit: a fit value of exponential
            b_fit: b fit value of exponential
            covariance: covariance matrix of a/b fit
            bit_threshold: bitscore threshold of dataset
        """
        distance, high, low = self._interpolate(distance, a_fit, b_fit, covariance)

        prediction = exponential(distance, a_fit, b_fit)
        self.axes.plot(distance, prediction, color="red", label="Predicted bitscore")
        self.axes.plot(distance, high, color="black")
        self.axes.plot(distance, low, color="black")

        self.axes.fill_between(
            distance,
            high,
            low,
            facecolor="blue",
            alpha=0.2,
            label="99% confidence interval",
        )

        self.interval = (max(distance) - min(distance)) / 100
        self.axes.set_xlim([-self.interval, max(distance) + self.interval])
        self.axes.set_ylim([0, max(prediction) * 1.1])
        self.axes.axhline(
            y=bit_threshold,
            linestyle="dashed",
            c="black",
            label="Detectability threshold",
        )

    @staticmethod
    def _interpolate(
        distance: npt.NDArray[np.float64],
        a_fit: float,
        b_fit: float,
        covariance: npt.NDArray[np.float64],
    ) -> tuple[npt.NDArray[np.float64], pd.Series[float], pd.Series[float]]:
        """Interpolate fits for smooth plotting.

        Args:
            distance: target x values
            a_fit: a fit value of exponential
            b_fit: b fit value of exponential
            covariance: covariance matrix of a/b fit

        Returns:
            new distance with 100 points
            high values of estimate
            low values of estimate
        """
        distance = np.linspace(distance.min(), distance.max(), num=100, endpoint=True)
        random = np.random.default_rng()
        result = find_confidence_interval(
            random,
            distance.reshape(-1, 1),
            sample_parameters(random, a_fit, b_fit, covariance),
        )

        return distance, result.high_interval, result.low_interval

    def set_axes(self, distance: list[float], species: pd.Index) -> None:
        """Set axes labels and ticks.

        Args:
            distance: target x values
            species: species names
        """
        self.axes.set_ylabel(
            "Bitscore",
            fontsize=13,
            labelpad=10,
        )

        self.axes.set_xlabel(
            "Evolutionary distance from focal species",
            fontsize=13,
            labelpad=10,
        )

        self.axes.spines["right"].set_visible(False)
        self.axes.spines["top"].set_visible(False)
        self.axes.tick_params(axis="x", width=2, length=7, direction="inout")
        self.axes.set_xticks(distance, species, fontsize=10, rotation=90)

        self.axes.tick_params(axis="y", labelsize=10)

    def highlight_not_in_fit(self, distance: list[float], in_fit: list[bool]) -> None:
        """Highlight species not included in fit.

        Args:
            distance: target x values
            in_fit: specify if each species was included in fit
        """
        for i, is_in_fit in enumerate(in_fit):
            if not is_in_fit:
                self.axes.axvspan(
                    distance[i] - self.interval,
                    distance[i] + self.interval,
                    facecolor=ORANGE,
                    alpha=0.3,
                    capstyle="round",
                )
                self.axes.get_xticklabels()[i].set_color(ORANGE)
                self.axes.get_xticklabels()[i].set_weight("bold")

    def highlight_ambiguous(self, ambiguous: list[bool]) -> None:
        """Mark species whose homology was ambiguous.

        Args:
            ambiguous: list of bools specifying if the species was ambiguous
        """
        for i, is_ambiguous in enumerate(ambiguous):
            if is_ambiguous:
                self.axes.get_xticklabels()[i].set_color(GREY)

    def show_label(self, any_not_in_fit: bool, any_ambiguous: bool) -> None:
        """Produce legend for the plot.

        Args:
            any_not_in_fit: true if any species were not present in fit
            any_ambiguous: true if any species were ambiguous
        """
        handles, labels = self.axes.get_legend_handles_labels()

        if any_ambiguous:
            handles.append(Patch(facecolor=GREY, alpha=0.3))
            labels.append("Homolog detected, but orthology ambiguous")

        if any_not_in_fit:
            handles.append(Patch(facecolor=ORANGE, alpha=0.3))
            labels.append("No homolog detected!")

        self.axes.legend(handles, labels, fontsize=9)

    def save(self, filename: str, format: str = "svg") -> None:
        """Save this figure to file.

        Args:
            filename: the file to save to
        """
        self.figure.savefig(filename, format=format)

    @staticmethod
    def show() -> None:
        """Show the plot."""
        plt.show()

    def generate_plot(self, result):
        self.title(result.gene, result.a_fit, result.b_fit, result.correlation)

        fit = result.result

        self.scores(fit.distance, fit.score, fit.dist_stdev)

        self.fit(
            fit.distance,
            a_fit=result.a_fit,
            b_fit=result.b_fit,
            covariance=result.covariance,
            bit_threshold=result.bit_threshold,
        )

        self.set_axes(fit.distance, fit.index)
        # test if the species is also ambiguous, that takes precedence
        self.highlight_not_in_fit(fit.distance, (fit.in_fit | fit.ambiguous))
        self.highlight_ambiguous(fit.ambiguous)
        self.show_label(
            any_not_in_fit=not (fit.in_fit | fit.ambiguous).all(),
            any_ambiguous=fit.ambiguous.any(),
        )

    def write_error(self, gene, error):
        self.axes.set_title(f"{gene}: ERROR")
        self.axes.text(0.5, 0.5, error, fontsize=30, horizontalalignment='center')



"""Collection of functions not tied to a particular object."""
from __future__ import annotations

import typing

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.stats


@typing.no_type_check
def exponential(domain, scale, decay):
    """Exponential function for fitting/plotting."""
    return scale * np.exp(-decay * domain)


@typing.no_type_check
def odr_exponential(parameters, domain):
    """Exponential function for odr estimates."""
    return parameters[0] * np.exp(-parameters[1] * domain)


def sample_parameters(
    random: np.random.Generator,
    a_fit: float,
    b_fit: float,
    covariance: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    function to, where possible, use maximum likelihood estimates of a and b
    parameter plus estimated covariance matrix to directly sample from the
    probability distribution of a and b (assume Gaussian with mean of max
    likelihood estimates and given covariance structure)
    """
    return random.multivariate_normal([a_fit, b_fit], covariance, size=200)


def find_confidence_interval(
    random: np.random.Generator,
    distances: npt.NDArray[np.float64],
    sampled_parameters: npt.NDArray[np.float64],
    bit_threshold: pd.DataFrame | None = None,
    index: pd.Index | None = None,
) -> pd.DataFrame:
    """Gives an empirical estimate of the prediction interval.
    function to take each of the sampled a, b values and use them to sample
    directly from the distribution of scores taking into account the Gaussian
    noise (a function of distance, a, b)."""
    # species x samples
    point_estimates = exponential(
        distances,
        sampled_parameters[:, 0],
        sampled_parameters[:, 1],
    )

    # species x samples
    exp = np.exp(-1 * sampled_parameters[:, 1] * distances)
    variance = sampled_parameters[:, 0] * (1 - exp) * exp

    # if variance is negative, need to ignore those samples
    invalid = variance <= 0
    # set to 0 to handle issues with sqrt and random normal
    variance[invalid] = 0

    # with normal, can't broadcast a shape on matrix inputs, instead
    # perform a standard normal draw and do rescaling manually
    # species x samples x draws
    draws = (
        random.standard_normal(size=(*point_estimates.shape, 200))
        * np.sqrt(variance[..., None])
        + point_estimates[..., None]
    )

    # set invalid values to nan, keep first entry as point estimate
    # this works because the invalid entries have 0 variance.
    draws[invalid, 1:] = np.nan

    # now find mean and std, ignoring nans
    mean = np.nanmean(draws, axis=(1, 2))  # type: ignore[no-untyped-call]
    std = np.nanstd(draws, axis=(1, 2))  # type: ignore[no-untyped-call]

    low, high = scipy.stats.norm.interval(0.99, mean, std)

    # calculate p values analytically from std estimate
    if bit_threshold is not None:
        p_values = scipy.stats.norm.cdf(bit_threshold["bit_threshold"], mean, std)

        return pd.DataFrame(
            {
                "p_values": p_values,
                "low_interval": low,
                "high_interval": high,
            },
            index=index,
        )

    return pd.DataFrame(
        {
            "low_interval": low,
            "high_interval": high,
        },
        index=index,
    )



from __future__ import annotations

import itertools
from io import StringIO

import pandas as pd
import pytest

from abSENSE.parameters import AbsenseParameters
from abSENSE.recorder import Recorder
from abSENSE.results import SampledResult


def test_open(faked_recorder):
    rec, files = faked_recorder

    assert len(files) == 0

    with rec.open() as _:
        assert len(files) == 6
        for file in files.values():
            assert file.closed is False

    for file in files.values():
        assert file.closed


def test_write_headers(faked_recorder):
    rec, files = faked_recorder
    with rec.open() as recorder:
        recorder.write_headers()

        header = "Gene\ts1\ts2\ts3\n"
        assert files["./predicted_bitscores.tsv"].getvalue() == (
            "# maximum likelihood bitscore predictions "
            "for each tested gene in each species\n" + header
        )
        assert files["./99PI_lower_prediction.tsv"].getvalue() == (
            "# the lower bound of the 99% bitscore prediction "
            "interval for each tested gene in each species\n" + header
        )
        assert files["./99PI_high_prediction.tsv"].getvalue() == (
            "# the upper bound of the 99% bitscore prediction "
            "interval for each tested gene in each species\n" + header
        )
        assert files["./failure_probabilities.tsv"].getvalue() == (
            "# the probability of a homolog being undetected "
            "at the specified significance threshold (see run info file) in each "
            "tested gene in each species\n" + header
        )
        assert files["./parameters.tsv"].getvalue() == (
            "# the best-fit parameters "
            "(performed using only bitscores from species not omitted from the fit; "
            "see run info file) for a and b for each gene\n"
            "Gene\ta\tb\n"  # not species header
        )
        assert files["./run_info.txt"].getvalue() == ""


@pytest.mark.parametrize("genelenfile", [None, "gene len file"])
@pytest.mark.parametrize("dblenfile", [None, "db len file"])
@pytest.mark.parametrize("pred_species", [[], ["a", "b", "c"], ["a"]])
def test_write_info(faked_recorder, genelenfile, dblenfile, pred_species):
    rec, files = faked_recorder

    include_only = None if pred_species == [] else ",".join(pred_species)
    dist = StringIO()
    dist.name = "distance file"
    score = StringIO()
    score.name = "score file"
    db_lens = None
    if dblenfile is not None:
        db_lens = StringIO()
        db_lens.name = dblenfile
    gen_lens = None
    if genelenfile is not None:
        gen_lens = StringIO()
        gen_lens.name = genelenfile
    score.name = "score file"
    params = AbsenseParameters(
        distances=dist,
        bitscores=score,
        e_value="e value",
        predict_all=False,
        plot_all=False,
        validate=False,
        plot="",
        include_only=include_only,
        gene_lengths=gen_lens,
        db_lengths=db_lens,
        out_dir=None,
        start_time="start time",
    )

    with rec.open() as recorder:
        recorder.write_info(params)

        for key, file in files.items():
            if key != "./run_info.txt":
                assert file.getvalue() == ""
                continue

            genelen = "Gene length (for all genes): 400 (default)\n"
            if genelenfile is not None:
                genelen = f"Gene length file: {genelenfile}\n"

            dblen = "Database length (for all species): 8000000 (default)\n"
            if dblenfile is not None:
                dblen = f"Database length file: {dblenfile}\n"

            pred = "Species used in fit: All (default)\n"
            if pred_species != []:
                pred = f"Species used in fit: {' '.join(pred_species)}\n"

            assert file.getvalue() == (
                "abSENSE analysis run on start time\n"
                "Input bitscore file: score file\n"
                "Input distance file: distance file\n"
                f"{genelen}"
                f"{dblen}"
                f"{pred}"
                "E-value threshold: e value\n"
            )


def test_write_gene(faked_recorder):
    rec, files = faked_recorder
    with rec.open() as recorder:
        recorder.write_gene("my gene name")
        for key, file in files.items():
            if key == "./run_info.txt":
                assert file.getvalue() == ""
                continue

            assert file.getvalue() == "my gene name"


def test_analysis_error(faked_recorder):
    rec, files = faked_recorder
    with rec.open() as recorder:
        recorder.analysis_error()
        string = "analysis_error"
        for key, file in files.items():
            if key == "./run_info.txt":
                assert file.getvalue() == ""
                continue

            if key == "./parameters.tsv":
                assert file.getvalue() == f"\t{string}\t{string}\n"
                continue

            assert file.getvalue() == f"\t{string}\t{string}\t{string}\n"


def test_not_enough_data(faked_recorder):
    rec, files = faked_recorder
    with rec.open() as recorder:
        recorder.not_enough_data()
        string = "not_enough_data"
        for key, file in files.items():
            if key == "./run_info.txt":
                assert file.getvalue() == ""
                continue

            if key == "./parameters.tsv":
                assert file.getvalue() == f"\t{string}\t{string}\n"
                continue

            assert file.getvalue() == f"\t{string}\t{string}\t{string}\n"


def old_write_result(
    mloutputfile,
    highboundoutputfile,
    lowboundoutputfile,
    pvaloutputfile,
    outputfileparams,
    predall,
    is_considered,
    is_ambiguous,
    realscores,
    a,
    b,
    predictions,
    highs,
    lows,
    pvals,
):
    for j in range(0, len(predictions)):
        prediction = predictions[j]
        lowprediction, highprediction, pval = (
            round(lows[j], 2),
            round(highs[j], 2),
            round(pvals[j], 2),
        )
        if not is_considered[j] and not is_ambiguous[j]:
            mloutputfile.write(f"\t{prediction}")
            highboundoutputfile.write(f"\t{highprediction}")
            lowboundoutputfile.write(f"\t{lowprediction}")
            pvaloutputfile.write(f"\t{pval}")
        elif is_considered[j]:
            if predall:
                realscore = realscores[j]
                mloutputfile.write(f"\t{prediction}(Ortholog_detected:{realscore})")
                highboundoutputfile.write(f"\t{highprediction}(Ortholog_detected)")
                lowboundoutputfile.write(f"\t{lowprediction}(Ortholog_detected)")
                pvaloutputfile.write(f"\t{pval}(Ortholog_detected)")
            else:
                mloutputfile.write("\tOrtholog_detected")
                highboundoutputfile.write("\tOrtholog_detected")
                lowboundoutputfile.write("\tOrtholog_detected")
                pvaloutputfile.write("\tOrtholog_detected")
        elif is_ambiguous[j]:
            if predall:
                mloutputfile.write(
                    f"\t{prediction}" "(Homolog_detected(orthology_ambiguous))"
                )
                highboundoutputfile.write(
                    f"\t{highprediction}" "(Homolog_detected(orthology_ambiguous))"
                )
                lowboundoutputfile.write(
                    f"\t{lowprediction}" "(Homolog_detected(orthology_ambiguous))"
                )
                pvaloutputfile.write(
                    f"\t{pval}" "(Homolog_detected(orthology_ambiguous))"
                )
            else:
                mloutputfile.write("\tHomolog_detected(orthology_ambiguous)")
                highboundoutputfile.write("\tHomolog_detected(orthology_ambiguous)")
                lowboundoutputfile.write("\tHomolog_detected(orthology_ambiguous)")
                pvaloutputfile.write("\tHomolog_detected(orthology_ambiguous)")

    mloutputfile.write("\n")
    highboundoutputfile.write("\n")
    lowboundoutputfile.write("\n")
    pvaloutputfile.write("\n")
    outputfileparams.write("\t" + str(a) + "\t" + str(b))
    outputfileparams.write("\n")


all_true_false = list(itertools.combinations_with_replacement([True, False], 3))
numbers = [[0.1, 0.2, 0.3], [-0.12345, 0.12345, 0.2468]]


@pytest.mark.parametrize("predall", [False, True])
@pytest.mark.parametrize("predictions", numbers)
@pytest.mark.parametrize("highs", numbers)
@pytest.mark.parametrize("lows", numbers)
@pytest.mark.parametrize("pvals", numbers)
@pytest.mark.parametrize("is_ambiguous", all_true_false)
@pytest.mark.parametrize("is_considered", all_true_false)
def test_write_result(
    faked_recorder,
    predall,
    highs,
    predictions,
    lows,
    pvals,
    is_ambiguous,
    is_considered,
):
    # get old results
    a = "a value"
    b = "b value"
    realscores = [0, 1, 2]
    test_files = [StringIO() for _ in range(5)]
    old_write_result(
        test_files[0],
        test_files[1],
        test_files[2],
        test_files[3],
        test_files[4],
        predall,
        is_considered,
        is_ambiguous,
        realscores,
        a,
        b,
        predictions,
        highs,
        lows,
        pvals,
    )

    rec, files = faked_recorder
    rec.predict_all = predall
    with rec.open() as recorder:
        for prediction, high, low, pval, realscore, considered, ambiguous in zip(
            predictions, highs, lows, pvals, realscores, is_considered, is_ambiguous
        ):
            recorder.write_result(
                prediction, high, low, pval, realscore, considered, ambiguous
            )
        recorder.write_params(a, b)
        recorder.finalize_row()

        assert files["./predicted_bitscores.tsv"].getvalue() == test_files[0].getvalue()
        assert (
            files["./99PI_high_prediction.tsv"].getvalue() == test_files[1].getvalue()
        )
        assert (
            files["./99PI_lower_prediction.tsv"].getvalue() == test_files[2].getvalue()
        )
        assert (
            files["./failure_probabilities.tsv"].getvalue() == test_files[3].getvalue()
        )
        assert files["./parameters.tsv"].getvalue() == test_files[4].getvalue()

    for file in test_files:
        file.close()


def test_plot(default_params, mocker):
    mocker.patch("abSENSE.recorder.os.makedirs")
    default_params.plot_all = True
    recorder = Recorder.build_recorder(default_params, [])
    result = StringIO(
        ",distance,score,p_values,low_interval,high_interval,ambiguous,in_fit,prediction,dist_stdev\n"
        "S_cer,0.0,2284.0,0.0,2260.2,2467.68,False,True,2359.91,1\n"
        "S_par,0.05,2254.0,0.0,2187.99,2378.63,False,True,2279.92,1\n"
        "S_mik,0.09,2234.0,0.0,2136.21,2316.95,False,True,2223.59,1\n"
        "S_kud,0.104,2197.0,0.0,2113.86,2291.22,False,True,2199.67,1\n"
        "S_bay,0.108,2200.0,0.0,2108.34,2284.71,False,True,2193.73,1\n"
        "S_castellii,0.36,1895.0,0.0,1765.16,1928.92,False,True,1846.31,1\n"
        "K_waltii,0.49,1767.0,0.0,1600.4,1778.84,False,True,1689.81,1\n"
        "A_gossyppi,0.52,1736.0,0.0,1571.01,1753.6,False,True,1662.61,1\n"
        "K_lactis,0.56,1691.0,0.0,1525.83,1711.85,False,True,1619.34,1\n"
        "A_nidulans,0.9,1099.0,0.0,1164.61,1395.01,False,True,1281.56,1\n"
        "S_pombe,0.92,1194.0,0.0,1147.89,1379.32,False,True,1265.21,1\n"
        "Y_lipolytica,0.95,1297.0,0.0,1119.85,1353.19,False,True,1238.13,1\n"
    )
    df = pd.read_csv(result, index_col=0)
    covariance = [[1.69e03, 1.20e00], [1.20e00, 1.82e-03]]

    fit = SampledResult(
        "GENE",
        df,
        2359.9123,
        0.6761,
        correlation=0.9,
        bit_threshold=[40],
        covariance=covariance,
    )
    result = StringIO()
    recorder.plot(fit, outfile=result)

    assert "DOCTYPE svg" in result.getvalue()


def test_plot_interpolated(default_params, mocker):
    mocker.patch("abSENSE.recorder.os.makedirs")
    default_params.plot_all = True
    recorder = Recorder.build_recorder(default_params, [])
    result = StringIO(
        ",distance,score,p_values,low_interval,high_interval,ambiguous,in_fit,prediction,dist_stdev\n"
        "S_cer,0.0,352.0,0.0,340.2397797315502,363.6119751420536,False,True,352.34734052045684,1\n"
        "S_par,0.051,171.0,2.659628935736003e-38,143.35192031605413,194.21790598475133,False,True,168.53197986670452,1\n"
        "S_mik,0.08800000000000001,92.8,5.933125751798271e-11,76.04084482029185,122.02297852984503,False,True,98.69974465645257,1\n"
        "S_kud,0.10400000000000001,82.0,3.993489447692749e-06,57.26044922851021,100.11382037310618,False,True,78.31290016903162,1\n"
        "S_bay,0.10800000000000001,0.0,2.8986806501042924e-05,53.27984272463878,95.12761609215814,False,False,73.91160397821581,1\n"
        "S_castellii,0.363,0.0,1.0,-1.6973045300440253,5.503161698002151,False,False,1.8504154754087632,1\n"
        "K_waltii,0.494,0.0,1.0,-1.1019801810161045,1.6744685716904901,False,False,0.2783344536876363,1\n"
        "A_gossyppi,0.518,0.0,1.0,-0.9622721726725029,1.3809148499246369,False,False,0.19671745018170525,1\n"
        "K_lactis,0.557,0.0,1.0,-0.7618769645703248,1.0033872194043867,False,False,0.11192220370314998,1\n"
        "A_nidulans,0.903,0.0,1.0,-0.0731876805054268,0.07511789426512828,False,False,0.0007515837789458042,1\n"
        "S_pombe,0.922,0.0,1.0,-0.06373870069810762,0.06514697376269933,False,False,0.0005710237430386822,1\n"
        "Y_lipolytica,0.9540000000000001,0.0,1.0,-0.05100568090545824,0.051966598919634975,False,False,0.000359491580422705,1\n"
    )
    df = pd.read_csv(result, index_col=0)
    covariance = [[26.43, 1.02], [1.02, 0.16]]

    fit = SampledResult(
        "NP_116682.3",
        df,
        352.347,
        14.4606,
        correlation=0.99,
        bit_threshold=[41.54],
        covariance=covariance,
    )

    result = StringIO()
    recorder.plot(fit, outfile=result)
    assert "DOCTYPE svg" in result.getvalue()



# pylint: disable=invalid-name

"""
Copyright (c) 2022 Troy Comi. All rights reserved.

abSENSE: Perform homolog detectability
"""
from __future__ import annotations

__version__ = "0.1.0"

__all__ = ("__version__",)



from __future__ import annotations

from io import StringIO

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

from abSENSE.analyzer import AbsenseAnalyzer
from abSENSE.exceptions import MissingGeneException, MissingSpeciesException
from abSENSE.recorder import Recorder
from abSENSE.results import (
    ErrorResult,
    NotEnoughDataResult,
    SampledResult,
    ValidationResult,
)
from abSENSE.utilities import find_confidence_interval, sample_parameters


def test_init_defaults(default_params):
    with pytest.warns(UserWarning) as record:
        analyzer = AbsenseAnalyzer(default_params)

    assert analyzer.odr_model is None

    assert len(record) == 1
    assert record[0].message.args[0] == (
        "Only one estimate of distances provided, "
        "additional values can improve result accuracy"
    )

    assert (analyzer.gene_lengths.index == analyzer.genes).all()
    assert (analyzer.gene_lengths["length"] == default_params.default_gene_length).all()

    assert (analyzer.db_lengths.index == analyzer.species).all()
    assert (analyzer.db_lengths["length"] == default_params.default_db_length).all()


def test_init_defaults_zero_variance(default_params):
    default_params.distances = StringIO(
        "S_cer\t0\t0\t0\t0\n"
        "S_par\t0.051\t0.051\t0.051\t0.051\n"  # zero variance
        "S_mik\t0.088\t0.098\t0.078\t0.098\n"
        "S_kud\t0.104\t0.114\t0.094\t0.114\n"
        "S_bay\t0.108\t0.118\t0.098\t0.107\n"
        "S_castellii\t0.363\t0.353\t0.373\t0.366\n"
        "K_waltii\t0.494\t0.484\t0.504\t0.496\n"
        "A_gossyppi\t0.518\t0.508\t0.528\t0.520\n"
        "K_lactis\t0.557\t0.547\t0.567\t0.559\n"
        "A_nidulans\t0.903\t0.893\t0.913\t0.905\n"
        "S_pombe\t0.922\t0.912\t0.932\t0.924\n"
        "Y_lipolytica\t0.954\t0.944\t0.964\t0.956\n"
    )
    with pytest.warns(UserWarning) as record:
        analyzer = AbsenseAnalyzer(default_params)

    assert analyzer.odr_model is None

    assert len(record) == 1
    assert record[0].message.args[0] == (
        "One or more estimates have 0 " "variance, skipping ODR"
    )

    assert (analyzer.gene_lengths.index == analyzer.genes).all()
    assert (analyzer.gene_lengths["length"] == default_params.default_gene_length).all()

    assert (analyzer.db_lengths.index == analyzer.species).all()
    assert (analyzer.db_lengths["length"] == default_params.default_db_length).all()


@pytest.mark.filterwarnings("ignore: Only one estimate of distances")
def test_init_with_include_only(default_params):
    # include only (or the distance index on default)
    # determines the order of bitscore columns
    default_params.include_only = "S_cer,S_mik,S_par"
    analyzer = AbsenseAnalyzer(default_params)
    assert (analyzer.species == "S_cer,S_mik,S_par".split(",")).all()
    assert (analyzer.bitscores.columns == "S_cer,S_mik,S_par".split(",")).all()


@pytest.mark.filterwarnings("ignore: Only one estimate of distances")
def test_init_with_include_only_missing(default_params):
    default_params.include_only = "S_cer,S_par,S_mik,NOTHERE"
    with pytest.raises(MissingSpeciesException) as error:
        AbsenseAnalyzer(default_params)
    assert str(error.value) == (
        "Unable to find all requested species in bitscores. " "Missing: ['NOTHERE']"
    )


@pytest.mark.filterwarnings("ignore: Only one estimate of distances")
def test_init_with_missing_db_species(default_params):
    default_params.include_only = "S_cer,S_par,S_mik,S_bay"
    default_params.db_lengths = StringIO(
        "#Species\tDBsize\n" "S_par\t2606124\n" "S_mik\t2603233\n" "S_cer\t2615464\n"
    )
    with pytest.raises(MissingSpeciesException) as error:
        AbsenseAnalyzer(default_params)
    assert str(error.value) == (
        "Unable to find all requested species in database lengths. "
        "Missing: ['S_bay']"
    )


@pytest.mark.filterwarnings("ignore: Only one estimate of distances")
def test_init_with_missing_gene_lengths(default_params):
    default_params.gene_lengths = StringIO(
        "#GeneID\tGeneLength\n"
        "NP_001018029.1\t66\n"
        "NP_001018030.1\t360\n"
        "NP_001018031.2\t113\n"
    )
    with pytest.raises(MissingGeneException) as error:
        AbsenseAnalyzer(default_params)
    assert str(error.value).startswith(
        "Unable to find all requested genes in gene lengths. "
        "Missing: ['NP_010181.2', "
    )


@pytest.fixture()
def analyzer_with_files(fungi_database_lengths, fungi_gene_lengths, default_params):
    default_params.gene_lengths = StringIO(fungi_gene_lengths)
    default_params.db_lenghts = StringIO(fungi_database_lengths)
    with pytest.warns(UserWarning):
        return AbsenseAnalyzer(default_params)


@pytest.fixture()
def analyzer_with_replicates(
    fungi_database_lengths,
    fungi_gene_lengths,
    quick_distances_replicates,
    default_params,
):
    default_params.distances = StringIO(quick_distances_replicates)
    default_params.distances.name = "distances"
    default_params.gene_lengths = StringIO(fungi_gene_lengths)
    default_params.db_lenghts = StringIO(fungi_database_lengths)
    return AbsenseAnalyzer(default_params)


def test_fit_gene_not_enough(analyzer_with_files):
    result = analyzer_with_files.fit_gene(
        "NP_001018030.1",
        pd.Series(
            [0, 1, 2] + [np.nan] * 9,
            index=analyzer_with_files.species,
        ),
    )
    assert result == NotEnoughDataResult("NP_001018030.1")


def test_fit_gene_analysis_error(analyzer_with_files, mocker):
    mocker.patch("scipy.optimize.curve_fit", side_effect=RuntimeError)
    result = analyzer_with_files.fit_gene(
        "NP_001018030.1",
        pd.Series(
            [3, 2, 1, 0] + [np.nan] * 8,
            index=analyzer_with_files.species,
        ),
    )
    assert result == ErrorResult("NP_001018030.1")


def test_fit_gene_infinite_covariance(analyzer_with_files, mocker):
    mocker.patch(
        "scipy.optimize.curve_fit", return_value=((1, 1), [[np.inf, 0], [0, 1]])
    )
    result = analyzer_with_files.fit_gene(
        "NP_001018030.1",
        pd.Series(
            [3, 2, 1, 0] + [np.nan] * 8,
            index=analyzer_with_files.species,
        ),
    )
    assert isinstance(result, ErrorResult)
    assert result.gene == "NP_001018030.1"
    assert_series_equal(
        result.predictions, np.exp(-1 * analyzer_with_files.distances)["distance"]
    )


def test_fit_normal(analyzer_with_files):
    result = next(analyzer_with_files.fit_genes())
    assert isinstance(result, SampledResult)
    assert result.gene == "NP_010181.2"
    assert result.a_fit == pytest.approx(2359.91, abs=1e-2)
    assert result.b_fit == pytest.approx(0.67612, abs=1e-4)
    assert result.correlation == pytest.approx(-0.976, abs=1e-2)
    assert result.bit_threshold == pytest.approx(39.55, abs=1e-2)


def test_fit_normal_replicates(analyzer_with_replicates):
    assert analyzer_with_replicates.odr_model is not None
    result = next(analyzer_with_replicates.fit_genes())
    assert isinstance(result, SampledResult)
    assert result.gene == "NP_010181.2"
    assert result.a_fit == pytest.approx(2424.33, abs=1e-2)
    assert result.b_fit == pytest.approx(0.74619, abs=1e-4)
    assert result.correlation == pytest.approx(-0.976, abs=1e-2)
    assert result.bit_threshold == pytest.approx(39.55, abs=1e-2)


def test_sample_parameters(analyzer_with_files):
    result = sample_parameters(analyzer_with_files.random, 1, 1, [[1, 0], [0, 1]])
    a_vals = result[:, 0]
    b_vals = result[:, 1]
    assert np.mean(a_vals) == pytest.approx(1, abs=1e-1)
    assert np.mean(b_vals) == pytest.approx(1, abs=1e-1)
    assert result.shape == (200, 2)


def test_bit_threshold(analyzer_with_files):
    old_result = -1 * np.log2(
        analyzer_with_files.e_value / (100 * analyzer_with_files.db_lengths)
    )
    old_result = old_result.rename(columns={"length": "bit_threshold"})
    result = analyzer_with_files.bit_threshold(100)
    assert_frame_equal(result, old_result)

    old_result = -1 * np.log2(
        analyzer_with_files.e_value / (234 * analyzer_with_files.db_lengths)
    )
    old_result = old_result.rename(columns={"length": "bit_threshold"})
    result = analyzer_with_files.bit_threshold(234)
    assert_frame_equal(result, old_result)


def test_find_confidence_interval(analyzer_with_files):
    bit_threshold = pd.DataFrame(
        {"bit_threshold": 40.1},
        index=analyzer_with_files.species,
    )
    result = find_confidence_interval(
        analyzer_with_files.random,
        np.expand_dims(analyzer_with_files.distances["distance"].to_numpy(), axis=1),
        np.array(
            [
                [40.1, 1],
                [40.2, 2],
                [40.3, 3],
                [40.4, -1],
            ]
        ),
        bit_threshold,
        analyzer_with_files.species,
    )

    cer = result.iloc[0, :]
    assert cer["p_values"] == pytest.approx(0.089856, abs=1e-4)
    assert cer["low_interval"] == pytest.approx(39.962014, abs=1e-4)
    assert cer["high_interval"] == pytest.approx(40.537986, abs=1e-4)


def test_validation(mocker, default_params, analyzer_with_files):
    analyzer_with_files.validate = True
    default_params.validate = True
    default_params.gene_lengths.name = "GENE"

    mocker.patch("abSENSE.recorder.os.makedirs")
    default_params.out_dir = "."
    recorder = Recorder.build_recorder(default_params, analyzer_with_files.species)
    # replace files with stringIO for easier testing
    files = {}

    def new_stringIO(*args, **kwargs):
        result = StringIO()
        result.close = lambda: None
        result.name = args[0]
        files[args[0]] = result
        return result

    mocker.patch("abSENSE.recorder.open", side_effect=new_stringIO)

    with recorder.open() as rec:
        rec.write_info(default_params)
        for result in analyzer_with_files.fit_genes():
            assert isinstance(result, ValidationResult)
            result.record_to(rec)

    assert files["./run_info.txt"].getvalue().split("\n")[-2] == (
        "Run in validation mode"
    )

    files["./validation.tsv"].seek(0)
    result = pd.read_csv(files["./validation.tsv"], sep="\t", comment="#", index_col=0)
    assert (result["mre"].isna()).sum() == 1  # just focal species
    assert pd.isna(result.iloc[0, -1])  # at first index, last column



"""Module for performing the abSENSE analysis."""
from __future__ import annotations

import warnings
from typing import Callable, Generator

import numpy as np
import pandas as pd
import scipy.odr
import scipy.optimize
import scipy.stats

from abSENSE.exceptions import MissingGeneException, MissingSpeciesException
from abSENSE.parameters import AbsenseParameters
from abSENSE.results import (
    ErrorResult,
    FitResult,
    NotEnoughDataResult,
    SampledResult,
    ValidationResult,
)
from abSENSE.utilities import (
    exponential,
    find_confidence_interval,
    odr_exponential,
    sample_parameters,
)


class AbsenseAnalyzer:
    """The analyzer object."""

    def __init__(self, params: AbsenseParameters, seed: int = 0):
        self.random = np.random.default_rng(seed)
        self.validate = params.validate
        self.e_value = params.e_value
        distances = pd.read_csv(
            params.distances,
            delimiter="\t",
            comment="#",
            header=None,
            index_col=0,
        )

        stdev = distances.to_numpy().std(axis=1)
        # the stdev of position 0 is always 0 as distances are relative to that
        # replace with the other minimum
        stdev[0] = stdev[1:].min()

        self.odr_model = None
        if distances.to_numpy().shape[1] == 1:
            warnings.warn(
                "Only one estimate of distances provided, additional "
                "values can improve result accuracy"
            )
        elif (stdev == 0).any():
            warnings.warn("One or more estimates have 0 variance, skipping ODR")
        else:
            self.odr_model = scipy.odr.Model(odr_exponential)

        self.distances = pd.DataFrame(
            data={
                "distance": distances.to_numpy().mean(axis=1),
                "dist_stdev": stdev,
            },
            index=distances.index.rename("specie"),
        )

        self.bitscores = pd.read_csv(
            params.bitscores,
            delimiter="\t",
            comment="#",
            index_col=0,
        )

        self.species = self.distances.index
        self.genes = self.bitscores.index

        if params.include_only is not None:
            self.species = pd.Index(params.include_only.split(","))

        if params.gene_lengths is None:
            self.gene_lengths = pd.DataFrame(
                {"length": params.default_gene_length},
                index=self.genes,
            )
        else:
            self.gene_lengths = pd.read_csv(
                params.gene_lengths,
                delimiter="\t",
                comment="#",
                names=np.array(["gene", "length"]),
                index_col=0,
            )

        if params.db_lengths is None:
            self.db_lengths = pd.DataFrame(
                {"length": params.default_db_length},
                index=self.species,
            )
        else:
            self.db_lengths = pd.read_csv(
                params.db_lengths,
                delimiter="\t",
                comment="#",
                names=np.array(["specie", "length"]),
                index_col=0,
            )

        self.validate_species()
        self.validate_genes()

        self.gene_filter: Callable[[str], bool] = lambda x: True
        if params.need_plots():
            self.gene_filter = params.plot_test()

    def validate_species(self) -> None:
        """Checks that the requested species are present in the bitscores and lengths."""
        # all species in species are in bitscores
        self.bitscores = self.bitscores.loc[
            :, self.bitscores.columns.isin(self.species)  # type: ignore[index]
        ]
        # species not found in column
        if not self.species.isin(self.bitscores.columns).all():
            missing = list(self.species[~self.species.isin(self.bitscores.columns)])  # type: ignore[index]
            raise MissingSpeciesException(
                "Unable to find all requested species in bitscores. "
                f"Missing: {missing}"
            )

        # reorder columns to match species
        self.bitscores = self.bitscores[self.species]

        if not self.species.isin(self.db_lengths.index).all():
            missing = list(self.species[~self.species.isin(self.db_lengths.index)])  # type: ignore[index]
            raise MissingSpeciesException(
                "Unable to find all requested species in database lengths. "
                f"Missing: {missing}"
            )

    def validate_genes(self) -> None:
        """Checks that the genes present in bitscores are found in gene lengths."""

        if not self.genes.isin(self.gene_lengths.index).all():
            missing = list(self.genes[~self.genes.isin(self.gene_lengths.index)])  # type: ignore[index]
            raise MissingGeneException(
                "Unable to find all requested genes in gene lengths. "
                f"Missing: {missing}"
            )

    def total_genes(self) -> int:
        """The total number of genes in this analysis.

        Returns:
            Number of genes to analyze
        """
        return sum(self.gene_filter(gene) for gene in self.bitscores.index)

    def fit_genes(self) -> Generator[FitResult, None, None]:
        """Fit all genes in analysis.

        Returns:
            Generator of FitResults for each gene
        """
        for gene, bitscore in self.bitscores.iterrows():
            if self.gene_filter(str(gene)):
                yield self.fit_gene(str(gene), bitscore)

    def fit_gene(self, gene: str, bitscore: pd.Series[float]) -> FitResult:
        """Fit a gene.

        If validating, will return a MultiFitResult.

        Args:
            gene: gene name
            bitscore: the bitscores for the gene
        """
        if self.validate:
            fits: dict[str, FitResult] = {}
            for species, value in bitscore[1:].iteritems():
                if value == 0 or np.isnan(value):
                    continue
                val_bitscore = bitscore.copy()
                val_bitscore[species] = 0  # set species of interest bitscore to 0
                fits[str(species)] = self._fit_gene(gene, val_bitscore)
            return ValidationResult(gene, fits, bitscore)

        return self._fit_gene(gene, bitscore)

    def _fit_gene(self, gene: str, bitscore: pd.Series[float]) -> FitResult:
        """Fit a gene helper function.

        Args:
            gene: gene name
            bitscore: the bitscores for the gene
        """
        data = self.distances.merge(
            bitscore.rename("score"),
            left_index=True,
            right_index=True,
        )

        ambiguous = data["score"].isna()
        in_fit = ~data["score"].isna() & (data["score"] != 0)

        if in_fit.sum() <= 2:
            return NotEnoughDataResult(gene)

        try:
            fit_result = scipy.optimize.curve_fit(
                exponential,
                data.loc[in_fit, "distance"],
                data.loc[in_fit, "score"],
                bounds=((-np.inf, 0), (np.inf, np.inf)),
            )
            (a_fit, b_fit), covariance = fit_result[:2]

            if self.odr_model is not None:
                estimate = scipy.odr.ODR(
                    scipy.odr.RealData(
                        data.loc[in_fit, "distance"],
                        data.loc[in_fit, "score"],
                        sx=data.loc[in_fit, "dist_stdev"],
                    ),
                    self.odr_model,
                    beta0=[a_fit, b_fit],
                ).run()
                a_fit, b_fit = estimate.beta
                covariance = estimate.cov_beta
        except (RuntimeError, scipy.optimize.OptimizeWarning):
            return ErrorResult(gene)

        predictions = exponential(self.distances, a_fit, b_fit)

        if np.isinf(covariance).any():
            return ErrorResult(gene, predictions=predictions["distance"])

        gene_length = float(self.gene_lengths.loc[gene])
        bit_threshold = self.bit_threshold(gene_length)
        global_threshold = bit_threshold.mean()["bit_threshold"]
        correlation = self.correlation(data.score)

        intervals = find_confidence_interval(
            self.random,
            np.expand_dims(self.distances["distance"].to_numpy(), axis=1),  # type: ignore[no-untyped-call]
            sample_parameters(self.random, a_fit, b_fit, covariance),
            bit_threshold,
            index=self.species,
        )

        result = data.join(intervals).assign(
            ambiguous=ambiguous,
            in_fit=in_fit,
            prediction=predictions["distance"],
        )

        return SampledResult(
            gene,
            result=result,
            a_fit=a_fit,
            b_fit=b_fit,
            bit_threshold=global_threshold,
            correlation=correlation,
            covariance=covariance,
        )

    def bit_threshold(self, gene_length: float) -> pd.DataFrame:
        """From the provided gene_length, estimate bit_thresholds for all species."""
        result = (
            self.db_lengths.apply(np.log2)
            + np.log2(gene_length)
            - np.log2(self.e_value)
        )
        return pd.DataFrame({"bit_threshold": result["length"]}, index=result.index)

    def correlation(self, scores: pd.Series[float]) -> float:
        """For the given distance/scores, estimate the correlation."""
        data = self.distances.assign(score=scores)
        data = data[data.score > 0]
        _, _, correlation, _, _ = scipy.stats.linregress(
            data.distance, np.log(data.score)
        )
        return float(correlation)



"""Custom exceptions for the package."""
from __future__ import annotations


class AbsenseException(Exception):
    """Base exception class."""


class MissingSpeciesException(AbsenseException):
    """A species is missing from the analysis."""


class MissingGeneException(AbsenseException):
    """A gene is missing from the analysis."""



from __future__ import annotations

from io import StringIO

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import abSENSE.__main__ as main


def test_defaults(faked_recorder_non_closing, default_params):
    _, files = faked_recorder_non_closing
    with pytest.warns(UserWarning) as record:
        main.perform_analysis(default_params)

    assert len(record) == 1
    assert record[0].message.args[0] == (
        "Only one estimate of distances provided, "
        "additional values can improve result accuracy"
    )

    files["./parameters.tsv"].seek(0)
    result = pd.read_csv(files["./parameters.tsv"], delimiter="\t", comment="#")
    expected = pd.read_csv(
        StringIO(
            "# the best-fit parameters (performed using only bitscores from species not omitted from the fit; see run info file) for a and b for each gene\n"
            "Gene\ta\tb\n"
            "NP_010181.2\t2359.910340705106\t0.676126045968598\n"
            "NP_009362.1\t1053.5985693169105\t0.4073235374987012\n"
            "NP_014555.1\t732.8505436789137\t0.5595110997139245\n"
            "NP_116682.3\t352.34734049379813\t14.460626046962682\n"
            "NP_011284.1\t303.0830341558144\t5.143646534570729\n"
            "NP_011878.1\t589.7752939170018\t2.729326889253197\n"
            "NP_013320.1\t1552.4815439696647\t5.120090782287544\n"
            "NP_014160.2\t968.5487558750334\t0.9669086263255712\n"
            "NP_014890.1\t383.70820281955355\t4.700780867702744\n"
        ),
        delimiter="\t",
        comment="#",
    )
    assert_frame_equal(result, expected)

    assert files["./predicted_bitscores.tsv"].getvalue() == (
        "# maximum likelihood bitscore predictions for each tested gene in each species\n"
        "Gene\tS_cer\tS_par\tS_mik\tS_kud\tS_bay\tS_castellii\tK_waltii\tA_gossyppi\tK_lactis\tA_nidulans\tS_pombe\tY_lipolytica\n"
        "NP_010181.2\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\n"
        "NP_009362.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\n"
        "NP_014555.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\n"
        "NP_116682.3\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t73.91\t1.85\t0.28\t0.2\t0.11\t0.0\t0.0\t0.0\n"
        "NP_011284.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t46.85\t23.88\t21.11\t17.27\t2.91\t2.64\t2.24\n"
        "NP_011878.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t218.98\t153.16\t143.44\t128.96\t50.16\t47.62\t43.64\n"
        "NP_013320.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t15.24\t13.83\t11.74\n"
        "NP_014160.2\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t600.73\tOrtholog_detected\t565.23\tOrtholog_detected\t397.15\tOrtholog_detected\n"
        "NP_014890.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t37.63\t33.61\t27.98\t5.5\t5.03\t4.33\n"
    )

    assert files["./failure_probabilities.tsv"].getvalue() == (
        "# the probability of a homolog being undetected at the specified significance threshold (see run info file) in each tested gene in each species\n"
        "Gene\tS_cer\tS_par\tS_mik\tS_kud\tS_bay\tS_castellii\tK_waltii\tA_gossyppi\tK_lactis\tA_nidulans\tS_pombe\tY_lipolytica\n"
        "NP_010181.2\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\n"
        "NP_009362.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\n"
        "NP_014555.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\tOrtholog_detected\tHomolog_detected(orthology_ambiguous)\n"
        "NP_116682.3\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t0.0\t1.0\t1.0\t1.0\t1.0\t1.0\t1.0\t1.0\n"
        "NP_011284.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t0.29\t0.98\t0.99\t1.0\t1.0\t1.0\t1.0\n"
        "NP_011878.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t0.0\t0.0\t0.0\t0.0\t0.23\t0.28\t0.39\n"
        "NP_013320.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t1.0\t1.0\t1.0\n"
        "NP_014160.2\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t0.0\tOrtholog_detected\t0.0\tOrtholog_detected\t0.0\tOrtholog_detected\n"
        "NP_014890.1\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\tOrtholog_detected\t0.55\t0.65\t0.8\t1.0\t1.0\t1.0\n"
    )

    assert files["./run_info.txt"].getvalue() == (
        "abSENSE analysis run on NOW\n"
        "Input bitscore file: bitscores\n"
        "Input distance file: distances\n"
        "Gene length (for all genes): 400 (default)\n"
        "Database length (for all species): 8000000 (default)\n"
        "Species used in fit: All (default)\n"
        "E-value threshold: 0.001\n"
    )



"""Main CLI entry point."""
from __future__ import annotations

from datetime import datetime
from typing import Any

import click

from abSENSE.analyzer import AbsenseAnalyzer
from abSENSE.parameters import AbsenseParameters
from abSENSE.recorder import Recorder


@click.command(help="A method to interpret undetected homologs")
@click.option(
    "--distances",
    type=click.File("rt"),
    required=True,
    help="tsv file containing pairwise evolutionary distances "
    "between focal species and each of the other species.",
)
@click.option(
    "--bitscores",
    type=click.File("rt"),
    required=True,
    help="tsv file containing bitscores between focal "
    "species gene and orthologs in other species.",
)
@click.option("--e-value", default=0.001, help="E-value threshold. Default 0.001.")
@click.option(
    "--include-only",
    type=str,
    help="Comma separated list of species to include in analysis.",
)
@click.option(
    "--gene-lengths",
    type=click.File("rt"),
    help="tsv file containing amino acid lengths of all genes in "
    "bitscores. Used to accurately calculate E-value threshold. "
    "Default is 400aa for all genes. "
    "Only large deviations will qualitatively affect results.",
)
@click.option(
    "--db-lengths",
    type=click.File("rt"),
    help="tsv file containing amino acid size of each species' database."
    "Used to accurately calculate E-value threshold. "
    "Default is 400aa/gene * 20,000 genes = 8000000 for all species, "
    "intended to be the size of an average proteome. "
    "Only large deviations will significantly affect results.",
)
@click.option(
    "--predict-all",
    is_flag=True,
    help="If set, predicts bitscores and P(detectable) of homologs "
    "in all species, including those in which homologs were "
    "detected. By default, will make predictions only for homologs "
    "that seem to be absent.",
)
@click.option(
    "--plot-all",
    is_flag=True,
    help="If set, will plot fitting results for each gene present in "
    "the bitscores file. By default, will not produce plots.",
)
@click.option(
    "--plot",
    default="",
    help="A comma separated list of genes to plot",
)
@click.option(
    "--validate",
    is_flag=True,
    help="If set, will remove each gene in the bitscore file and tabulate "
    "statistics on how well the current distances fit the data.",
)
@click.option(
    "--out-dir",
    type=click.Path(file_okay=False),
    help="Name of output directory. " "Default is date and time when analysis was run.",
)
def main(**args: Any) -> None:
    """CLI wrapper."""
    now = datetime.now()
    start_time = now.strftime("%m.%d.%Y_%H.%M")

    params = AbsenseParameters(**args, start_time=start_time)

    perform_analysis(params)


def perform_analysis(params: AbsenseParameters) -> None:
    """Wrapper for unit testing."""
    analyzer = AbsenseAnalyzer(params)
    total_genes = analyzer.total_genes()

    with Recorder.build_recorder(params, analyzer.species).open() as recorder:
        recorder.write_info(params)
        recorder.write_headers()

        click.echo("Running!")

        for i, result in enumerate(analyzer.fit_genes()):
            status = result.status
            if status != "":
                status = "\t-- " + status
            click.echo(f"gene {i+1} out of {total_genes}: {result.gene}{status}")

            result.record_to(recorder)


if __name__ == "__main__":
    main()



"""Recorder module for handling conversion and recording of absense analyses."""
from __future__ import annotations

import os
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator, TextIO

import numpy as np
import pandas as pd

from abSENSE.parameters import AbsenseParameters
from abSENSE.plotting import FitPlot

if TYPE_CHECKING:
    from abSENSE.results import SampledResult


class Recorder(ABC):
    """Base class of an object for recording results."""

    def __init__(self, params: AbsenseParameters, species: pd.Index):
        os.makedirs(params.output_directory, exist_ok=True)
        self.output_dir = params.output_directory
        self.species = species
        self._info_file: TextIO | None = None

    @abstractmethod
    @contextmanager
    def open(self) -> Generator[Recorder, None, None]:
        """Context managed resources."""

    @abstractmethod
    def write_headers(self) -> None:
        """Called once prior to streaming data, setup headers."""

    def write_info(self, params: AbsenseParameters) -> None:
        """Write the run information to the info file."""
        if self._info_file is None:
            return

        self._info_file.write(f"abSENSE analysis run on {params.start_time}\n")
        self._info_file.write(f"Input bitscore file: {params.bitscores.name}\n")
        self._info_file.write(f"Input distance file: {params.distances.name}\n")

        if params.gene_lengths is None:
            self._info_file.write(
                f"Gene length (for all genes): {params.default_gene_length} (default)\n"
            )
        else:
            self._info_file.write(f"Gene length file: {params.gene_lengths.name}\n")

        self._info_file.write("Database length ")
        if params.db_lengths is None:
            self._info_file.write(
                f"(for all species): {params.default_db_length} (default)\n"
            )
        else:
            self._info_file.write(f"file: {params.db_lengths.name}\n")

        self._info_file.write("Species used in fit: ")
        if params.include_only is None:
            self._info_file.write("All (default)")
        else:
            self._info_file.write(" ".join(params.include_only.split(",")))
        self._info_file.write("\n")

        self._info_file.write(f"E-value threshold: {params.e_value}\n")

    @staticmethod
    def build_recorder(params: AbsenseParameters, species: pd.Index) -> Recorder:
        """Factory method to return Recorder subtype based on parameters."""
        if params.validate:
            return ValidationRecorder(params, species)
        return FileRecorder(params, species)


class FileRecorder(Recorder):
    """Contains file handles to record absense analysis results as text files."""

    def __init__(self, params: AbsenseParameters, species: pd.Index):
        super().__init__(params, species)
        self.predict_all = params.predict_all
        self.should_plot = params.plot_test()
        if params.need_plots():
            os.makedirs(f"{self.output_dir}/plots", exist_ok=True)
        self.filenames = {
            "bitscores": "predicted_bitscores.tsv",
            "low": "99PI_lower_prediction.tsv",
            "high": "99PI_high_prediction.tsv",
            "failure": "failure_probabilities.tsv",
            "params": "parameters.tsv",
        }
        self._files: dict[str, TextIO] = {}

    @contextmanager
    def open(self) -> Generator[Recorder, None, None]:
        """Context managed file opening."""
        for key, file in self.filenames.items():
            self._files[key] = open(f"{self.output_dir}/{file}", "w", encoding="utf8")
        self._info_file = open(f"{self.output_dir}/run_info.txt", "w", encoding="utf8")

        try:
            yield self

        finally:
            for handle in self._files.values():
                handle.close()
            self._info_file.close()

    def write_headers(self) -> None:
        """Write brief summary header for each file."""
        species_header = "Gene\t" + "\t".join(self.species) + "\n"
        self._files["bitscores"].write(
            "# maximum likelihood bitscore predictions "
            "for each tested gene in each species\n" + species_header
        )

        self._files["low"].write(
            "# the lower bound of the 99% bitscore prediction "
            "interval for each tested gene in each species\n" + species_header
        )

        self._files["high"].write(
            "# the upper bound of the 99% bitscore prediction "
            "interval for each tested gene in each species\n" + species_header
        )

        self._files["failure"].write(
            "# the probability of a homolog being undetected "
            "at the specified significance threshold (see run info file) in each "
            "tested gene in each species\n" + species_header
        )

        self._files["params"].write(
            "# the best-fit parameters "
            "(performed using only bitscores from species not omitted from the fit; "
            "see run info file) for a and b for each gene\n"
            "Gene\ta\tb\n"  # not species header
        )

    def write_gene(self, gene: str) -> None:
        """Record the gene column for all files."""
        for file in self._files.values():
            file.write(gene)

    def analysis_error(self, predictions: list[float] | None = None) -> None:
        """Record an analysis error for this gene."""
        self._write_str("analysis_error", predictions)

    def not_enough_data(self) -> None:
        """Record the gene does not have enough data."""
        self._write_str("not_enough_data")

    def _write_str(
        self,
        entry: str,
        bitscore_overrides: list[float] | None = None,
    ) -> None:
        """Write the entry for each column in this row."""
        line = "\t" + "\t".join([entry] * len(self.species)) + "\n"
        for key, file in self._files.items():
            if key == "params":
                file.write(f"\t{entry}\t{entry}\n")
            elif bitscore_overrides is not None and key == "bitscores":
                file.write(
                    "\t"
                    + "\t".join(str(round(val, 2)) for val in bitscore_overrides)
                    + "\n"
                )
            else:
                file.write(line)

    def write_result(
        self,
        prediction: float,
        high: float,
        low: float,
        pval: float,
        realscore: float,
        is_considered: bool,
        is_ambiguous: bool,
    ) -> None:
        """Record the fit values for this gene, species."""

        high = round(high, 2)
        low = round(low, 2)
        pval = round(pval, 2)

        site_type = ""
        additional_score = ""

        if is_considered:
            site_type = "Ortholog_detected"
            additional_score = f":{realscore}"
        elif is_ambiguous:
            site_type = "Homolog_detected(orthology_ambiguous)"

        if site_type == "":
            self._files["bitscores"].write(f"\t{prediction}")
            self._files["high"].write(f"\t{high}")
            self._files["low"].write(f"\t{low}")
            self._files["failure"].write(f"\t{pval}")

        else:
            if self.predict_all:
                self._files["bitscores"].write(
                    f"\t{prediction}({site_type}{additional_score})"
                )
                self._files["high"].write(f"\t{high}({site_type})")
                self._files["low"].write(f"\t{low}({site_type})")
                self._files["failure"].write(f"\t{pval}({site_type})")
            else:
                for key, file in self._files.items():
                    if key != "params":
                        file.write(f"\t{site_type}")

    def write_params(
        self,
        a_prediction: float,
        b_prediction: float,
    ) -> None:
        """Record the param values for this gene."""
        self._files["params"].write(f"\t{a_prediction}\t{b_prediction}")

    def finalize_row(self) -> None:
        """Write newlines to all files."""
        for file in self._files.values():
            file.write("\n")

    def plot(
        self,
        result: SampledResult,
        outfile: str | None = None,
    ) -> None:
        """Produce a plot from the provided SampledResult.

        Args:
            result: the sampled result
            outfile: if set, will save to the specified file.  Default to result.gene
        """
        if not self.should_plot(result.gene):
            return

        if outfile is None:
            outfile = f"{self.output_dir}/plots/{result.gene}.svg"

        plot = FitPlot()

        plot.generate_plot(result)

        plot.save(outfile)


class ValidationRecorder(Recorder):
    """Contains file handles to record absense analysis results as text files."""

    def __init__(self, params: AbsenseParameters, species: pd.Index):
        super().__init__(params, species)
        self.species = species
        self.results = pd.DataFrame(
            np.zeros((len(self.species), 3)),
            index=self.species,
            columns=["count", "mre", "p>0.05"],
        )

        self.p_values = pd.Series(
            "",
            index=self.species,
        )
        self._result_file: None | TextIO = None

    @contextmanager
    def open(self) -> Generator[Recorder, None, None]:
        """Context managed file opening."""
        self._info_file = open(f"{self.output_dir}/run_info.txt", "w", encoding="utf8")
        self._result_file = open(
            f"{self.output_dir}/validation.tsv", "w", encoding="utf8"
        )

        try:
            yield self
            # on return, write output
            self._result_file.write(
                "# count: number of valid estimates\n"
                "# mre: mean, percent relative error\n"
                "# p>0.05: number of estimates with p(not detectable) > 0.05\n"
                "# p_values: comma separated list of p(not detectable)\n"
            )

            self.results["mre"] = self.results.apply(
                lambda x: x["mre"] / x["count"] * 100 if x["count"] != 0 else np.nan,
                axis=1,
            )
            # remove any repeated commas (\3) or at front (\1) or back (\2)
            self.results["p_values"] = self.p_values.str.replace(
                r"(^,+)|(,+$)|(,)+", r"\3", regex=True
            )
            # loc maintains original order
            self.results.loc[self.species].to_csv(
                self._result_file,
                sep="\t",
                float_format="%.2f",
                columns=["count", "mre", "p>0.05", "p_values"],
            )

        finally:
            self._info_file.close()
            self._result_file.close()

    def write_info(self, params: AbsenseParameters) -> None:
        """Write the run information to the info file."""
        super().write_info(params)
        if self._info_file is None:
            return
        self._info_file.write("Run in validation mode\n")

    def write_headers(self) -> None:
        pass

    def aggregate(self, result: pd.DataFrame) -> None:
        """Record the result from a single gene to this validation recorder."""
        self.results = self.results.combine(
            result.loc[:, result.columns != "p_values"],  # type: ignore[index]
            np.add,
            fill_value=0,
        )

        # fails if no results were SampledResults
        if "p_values" in result:
            self.p_values = self.p_values.combine(
                result["p_values"],
                lambda a, b: str(a) + str(b) + ",",
                fill_value="",
            )



"""Helper methods for performing analyses with pyodide."""

import base64
from io import BytesIO, StringIO
from typing import TextIO
import pandas as pd

from abSENSE.analyzer import AbsenseAnalyzer
from abSENSE.parameters import AbsenseParameters
from abSENSE.plotting import FitPlot
from abSENSE.results import SampledResult, ErrorResult, NotEnoughDataResult

def get_analyzer_from_text(
        data: str,
        e_value: float,
        gene_length: float,
        db_length: float,
    ):
    """Given text input, return analyzer."""
    df = pd.read_csv(StringIO(data), sep=r'[\t|,]', engine='python')

    distances = StringIO()
    df.iloc[:, 0:2].to_csv(distances, sep='\t', header=False, index=False)
    distances.seek(0)

    bitscores = StringIO()
    bits = df.iloc[:, 2:].T
    bits.columns=df.iloc[:, 0]
    bits.index.name='Gene'
    bits.to_csv(bitscores, sep='\t', header=True, index=True, na_rep='N/A')
    bitscores.seek(0)

    return get_analyzer_from_files(
        bitscores=bitscores,
        distances=distances,
        e_value=e_value,
        gene_length=gene_length,
        db_length=db_length,
    )

def get_analyzer_from_files(
        bitscores: TextIO,
        distances: TextIO,
        e_value: float,
        gene_length: float,
        db_length: float,
    ):
    """Given file inputs, return analyzer."""
    params = AbsenseParameters(
        bitscores=bitscores,
        distances=distances,
        e_value=e_value,
        default_gene_length=gene_length,
        default_db_length=db_length,
        include_only=None,
        out_dir='pyodide',
        plot='',
        plot_all=True,
        predict_all=True,
        start_time='',
        validate=False,
        # these are expected to be files
        gene_lengths=None,
        db_lengths=None,
    )
    return AbsenseAnalyzer(params)

def get_plot(gene: str, analyzer: AbsenseAnalyzer):
    bitscore = analyzer.bitscores.loc[gene]
    result = analyzer.fit_gene(gene, bitscore)
    plot = FitPlot()
    if isinstance(result, SampledResult):
        plot.generate_plot(result)
    else:
        if isinstance(result, ErrorResult):
            plot.write_error(result.gene, "Analysis error.")
        elif isinstance(result, NotEnoughDataResult):
            plot.write_error(result.gene, "Not enough data.")

    return plot


def generate_png_string(plot: FitPlot):
    buf = BytesIO()
    plot.save(buf, format='png')
    buf.seek(0)
    base64_str = base64.b64encode(buf.read()).decode('UTF-8')
    img_str = f"data:image/png;base64,{base64_str}"



"""Result objects to represent various outcomes from analysis."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd

from abSENSE.recorder import FileRecorder, Recorder, ValidationRecorder


class FitResult:
    """Base class for fitting results."""

    def __init__(self, gene: str):
        self.gene = gene
        self.status = ""

    def __eq__(self, other: object) -> bool:
        """Test if this fitting result is the same as other."""
        return self.__dict__ == other.__dict__

    def record_to(self, recorder: Recorder) -> None:
        """Record this result to the Recorder.

        Args:
            recorder: a Recorder
        """
        if not isinstance(recorder, FileRecorder):
            raise ValueError()
        recorder.write_gene(self.gene)


class ErrorResult(FitResult):
    """An analysis result when an analysis error occurs."""

    def __init__(self, gene: str, predictions: list[float] | None = None):
        super().__init__(gene)
        self.predictions = predictions
        self.status = "Analysis Error"

    def record_to(self, recorder: Recorder) -> None:
        super().record_to(recorder)
        if not isinstance(recorder, FileRecorder):
            raise ValueError()
        recorder.analysis_error(predictions=self.predictions)


class NotEnoughDataResult(FitResult):
    """Analysis result when not enough data is present."""

    def __init__(self, gene: str):
        super().__init__(gene)
        self.status = "Not Enough Data"

    def record_to(self, recorder: Recorder) -> None:
        super().record_to(recorder)
        if not isinstance(recorder, FileRecorder):
            raise ValueError()
        recorder.not_enough_data()


class SampledResult(FitResult):
    """Analysis result when the fit is successful."""

    def __init__(
        self,
        gene: str,
        result: pd.DataFrame,
        a_fit: float,
        b_fit: float,
        bit_threshold: float,
        correlation: float,
        covariance: npt.NDArray[np.float64],
    ):
        super().__init__(gene)
        self.result = result
        self.a_fit = a_fit
        self.b_fit = b_fit
        self.bit_threshold = bit_threshold
        self.correlation = correlation
        self.covariance = covariance

    def record_to(self, recorder: Recorder) -> None:
        super().record_to(recorder)
        if not isinstance(recorder, FileRecorder):
            raise ValueError()
        recorder.plot(result=self)

        recorder.write_params(self.a_fit, self.b_fit)

        for _, row in self.result.round(2).iterrows():
            recorder.write_result(
                prediction=row.prediction,
                high=row.high_interval,
                low=row.low_interval,
                pval=row.p_values,
                realscore=row.score,
                is_considered=row.in_fit,
                is_ambiguous=row.ambiguous,
            )

        recorder.finalize_row()


class ValidationResult(FitResult):
    """Analysis result of iterative validation."""

    def __init__(
        self, gene: str, fits: dict[str, FitResult], bitscore: pd.Series[float]
    ):
        super().__init__(gene)
        # key is species which was set to 0, value is resulting fit
        self.fits = fits
        self.bitscore = bitscore

    def record_to(self, recorder: Recorder) -> None:
        if not isinstance(recorder, ValidationRecorder):
            raise ValueError()
        result = self._analyze_fits()
        recorder.aggregate(result)

    def _analyze_fits(self) -> pd.DataFrame:
        stats = {}
        for species, result in self.fits.items():
            if not isinstance(result, SampledResult):
                continue

            true_value = self.bitscore[species]
            statistics = result.result.loc[species]

            stats[species] = {
                "count": 1,
                "p>0.05": 1 if statistics["p_values"] > 0.05 else 0,
                "p_values": f'{statistics["p_values"]:.4f}',
                "mre": (statistics["prediction"] - true_value) / true_value,
            }
        return pd.DataFrame(stats).transpose()
